/*************************************************************************/
/*<EXP>*
SERIALIO - Primitive to send/receive commands to serial port device

<A:PORTNAME> - serial device name (eg "/dev/ttyS0")
<A:STR_TX>   - string to send (usually requires quotes "")
<R:RES_RX>   - name of result to hold received string from serial device

SWITCHES:
  /DEBUG    - debug info level
  /CRLF     - CR/LF mode (1=CR_Only,2=LF_Only,3=CR/LF,4=LF/CR)
  /IODELAY= - specify delay (uSec) between successive read/writes to port

WIDGETS:
  NONE

EXAMPLES:
  serialio/crlf=3 "/dev/ttyS0" "FRQ?" a:mystatus

NOTES:
  make certain the device read/write permissions allow user I/O:

            i.e. --- "chmod ogu+rw /dev/ttyS0"

  before using this primitive

  REVISION     PROGRAMMER      DESCRIPTION
  --------     ----------      -------------------------------------
  20071101     rdh             baseline

/*<EXP>*/
/*************************************************************************/

// Includes
#include <primitive.h>
#include <libserial.h>

// Begin: Main Routine
void mainroutine () 
{
  // Variable declarations
  void *SP; 
  string cpp_devname = "";
  string cpp_tx = "";
  string cpp_rx = "";
  string cpp_resname = "";
  char   c_tx[MAX_SERIALIO_STRINGLEN];
  char   c_rx[MAX_SERIALIO_STRINGLEN];
  bool_4 log_resname;
  int_4  i,initstat;             
  int_4  sw_debug, sw_crlf, sw_iodelay;

  // Switches
  sw_debug   = m_get_switch("DEBUG");
  sw_crlf    = m_get_switch("CRLF");    
  sw_crlf    = (sw_crlf > 0) ? (sw_crlf) : (3);
  sw_iodelay = m_get_switch("IODELAY");
  sw_iodelay = (sw_iodelay > 0) ? (sw_iodelay) : (75000); 

  // Command line
  cpp_devname = m_apick(1);
  cpp_tx      = m_apick(2);
  cpp_resname = m_upick(3);
  log_resname = m_ppick(3);

  // Sync and run
  m_sync();

  // Initialize "C" strings
  c_tx[0] = 0x00;
  c_rx[0] = 0x00;

  // Create serial port handle and initialize terminal
  serialport_create(&SP);
  serialport_set_debuglevel(SP,sw_debug);
  serialport_set_iodelay(SP,sw_iodelay);
  initstat = serialport_init(SP,cpp_devname.c_str());

  if (sw_debug > 0) {
    printf("serialport_get_debuglevel: %i\n",serialport_get_debuglevel(SP));
    printf("serialport_get_iodelay:    %i\n",serialport_get_iodelay(SP));
  }

  // Terminate the input TX string according to serial device specs
  strncpy(c_tx, cpp_tx.c_str(), MAX_SERIALIO_STRINGLEN);
  i = strnlen(c_tx, MAX_SERIALIO_STRINGLEN);
  i = (i < (MAX_SERIALIO_STRINGLEN-2)) ? (i) : (MAX_SERIALIO_STRINGLEN-2);
  switch (sw_crlf) {
    case (1):
      c_tx[i]   = 0x0d;  // <CR> only
      c_tx[i+1] = 0x00;  // NULL       
      break;
    case (2):
      c_tx[i]   = 0x0a;  // <LF> only
      c_tx[i+1] = 0x00;  // NULL       
      break;
    case (3):
      c_tx[i]   = 0x0d;  // <CR>
      c_tx[i+1] = 0x0a;  // <LF> 
      c_tx[i+2] = 0x00;  // NULL       
      break;
    case (4):
      c_tx[i]   = 0x0a;  // <LF>
      c_tx[i+1] = 0x0d;  // <CR> 
      c_tx[i+2] = 0x00;  // NULL       
      break;
    default:
      c_tx[i]   = 0x0d;  // <CR>
      c_tx[i+1] = 0x0a;  // <LF> 
      c_tx[i+2] = '\0';  // NULL       
  }  

  // Send the command and check for status
  if (initstat == 0) {
    serialport_write(SP,c_tx);
    serialport_read(SP,c_rx,MAX_SERIALIO_STRINGLEN);
  }

  // Release serial port and destroy handle
  serialport_destroy(&SP);

  // Output the result
  cpp_rx = string(c_rx,strnlen(c_rx,MAX_SERIALIO_STRINGLEN));
  if (sw_debug > 0) printf("Returned: %s\n",cpp_rx.c_str());
  if (log_resname) m_arslt(cpp_resname,cpp_rx);

}
