/********************************************************************/
/*<EXP>*
SOURCENIC - Gets data from multicast source (i.e. rumeL box) from UDP port 
            and sends data to file/pipe

<OUTPUT>    Output pipe  (header = header filled by m_initialize with SB data)
<ADDRESS>   Multicast address to receive data from (numerical in quotes)
<INTERFACE> IP address of interface to receive multicast data into
            (numerical in quotes)

Reads from the UDP port are blocking.  All data appearing at <PORT> will be
acquired, there is no check for which source it's coming from.

Socket buffer size is set internally to try to get /SBUFSZ, however, 
for RHEL4U2 the max is set in the kernel to 256k. You can change the max value
to something more reasonable for high speed acquisition with the following
command:

/sbin/sysctl -w net.core.rmem_max=1048576

You can check to see what the current value is with the following command:
/sbin/sysctl net.core.rmem_max

(this sets it to 2MBytes). I have demonstrated the above value to be
adequate for gigabit ethernet acquisition WITHOUT anything else running on
the same CPU. 

SWITCHES:
  /DEBUG      - debug info level
  /SDDSMODE=n - 1=raw_packets, 2=data_only, 3=header_only
  /FC=        - specify header format code as SB,SI,CB,CI (SB default)
  /XD=        - specify header abscissa increment (default 1.0)
  /PORT=n     - UDP port # to receive data (default = 29495, the SDDS port)
  /PRANGE=n   - range of ports to use starting at <PORT> (i.e. if
                 requested port is unavaliable)
  /TL=n       - # of bytes to transfer each pass. For SDDS data this
                 must be set to 1080 (the default)
  /SBUFSZ=n   - Desired socket buffer size in bytes.  Default=10M.
                 Kernel may or may not allow this size (see notes above)
  /NUMPKTS=   - Exit after output of specified number of SDDS packets


CONTROL WIDGETS:
  1. L:exit   - setting to any value other than 0 forces primitive to exit
  2. D:pktcnt - received packet counter
       
REVISION       PROGRAMMER      DESCRIPTION
--------       ----------      -------------------------------------
07-Feb-2007    dkc             baseline
27-Nov-2007    rdh             added output modes/switches/widgets

/*<EXP>*/
/********************************************************************/

// Includes
#include <primitive.h>
#include <fintrinsics.h>
#include <math.h>
#include "sourcenic.h"

// Begin: Main Routine
void mainroutine () 
{

  // Variable Declarations
  CPHEADER HOUT;
  int_4  socket;      // Socket ID as returned by opensocket routine
  int_4  sw_debug,sw_prange,sw_tl,sw_mode,sw_sbufsz,sw_socketport;
  int_1 *pack_buf;
  int_1 *mode_buf;
  int_4  mode_bytes, mode_elem;
  int_4  status;
  int_4  idw_exit, idw_pktcnt, wid_exit;
  real_8 pktctr,pktctr_max;
  real_8 sw_xd;
  string csw_fc = "SB";
  string multicast_address_label;
  string interface_address_label;

  // Get command line switches
    sw_debug      = m_get_switch("DEBUG");
    sw_tl         = m_get_switch_def("TL",1080);
    sw_sbufsz     = m_get_switch_def("SBUFSZ",1024*1024*10);
    sw_mode       = m_get_switch("SDDSMODE");
    sw_mode       = (sw_mode > 0) ? (sw_mode) : (1);
    sw_prange     = m_get_switch_def("PRANGE",1);
    sw_socketport = m_get_switch("PORT");
    sw_socketport = (sw_socketport > 0) ? (sw_socketport) : (29495);
    pktctr_max    = m_get_dswitch("NUMPKTS");
    sw_xd         = m_get_dswitch("XD");
    sw_xd         = (sw_xd > 0) ? (sw_xd) : (1.0);
    m_get_uswitch("FC",csw_fc);
    m_uppercase(csw_fc);
    
  // Get output file name
    m_initialize(HOUT);
      HOUT.file_name  = m_apick(1);
      HOUT.type       = 1000;
      HOUT.format     = csw_fc;
      HOUT.xdelta     = sw_xd;
    m_allocate(HOUT);
  
  // Get multicast address and interface to use
    multicast_address_label = m_apick(2);
    interface_address_label = m_apick(3);

  // Allocate memory for packet buffer and define offsets
    pack_buf = (int_1 *)malloc((size_t)sw_tl);
    switch (sw_mode) {
      case (1):                    // SDDS_Header+Data
        mode_buf   = &pack_buf[0];             
        mode_bytes = 1080;
        mode_elem  = mode_bytes/(HOUT.bpe);
        break;
      case (2):                    // Data only
        mode_buf   = &pack_buf[56];             
        mode_bytes = 1024;
        mode_elem  = mode_bytes/(HOUT.bpe);
        break;
      case (3):                    // SDDS_Header only
        mode_buf   = &pack_buf[0];             
        mode_bytes = 56;   
        mode_elem  = mode_bytes;
        break;
      default:                     // SDDS_Header+Data
        mode_buf   = &pack_buf[0];             
        mode_bytes = 1080;
        mode_elem  = mode_bytes/(HOUT.bpe);
    }

  // Open UDP socket w/ multicast address
    socket = opensocket(&sw_socketport,sw_prange,
                        (char *)multicast_address_label.c_str(),
                        (char *)interface_address_label.c_str(), 
                        sw_sbufsz,sw_debug);
                        
    if (socket <= 0) 
      m_error("Opening socket");
    else if (Mu->verbose) 
      printf("SOURCENIC: Opened UDP port %d \n",sw_socketport);
   
 
    if (sw_debug > 0) {
      cout << "---- SOURCENIC Debug-----" << endl;
      cout << "Socket = " << socket << endl;
      cout << "Socket Port = " << sw_socketport << endl;
      cout << "Multicast Address = " << multicast_address_label << endl;
      cout << "Interface Address = " << interface_address_label << endl;
      cout << "Transfer Length = " << sw_tl << endl;
      cout << "Format Code     = " << csw_fc << endl;
      cout << "Mode            = " << sw_mode << endl;
      cout << "Mode_Bytes      = " << mode_bytes << endl;
      cout << "Mode_Elements   = " << mode_elem << endl;
    }
  
    // Widget status/controls
    wid_exit   = 0;
    idw_exit   = m_lwinit("Exit",wid_exit,0,1,1);
    idw_pktcnt = m_dwinit("PktCnt",0.0,0.0,1.0e30,1.0);  
    pktctr = 0;
    fflush(NULL);  //  forces a write of all user-space buffered data for
                   //  all open output streams
    
    m_sync();
    while ((wid_exit == 0)&&(! Mc->break_ )) {
     
      // Check control widgets
      m_lwget(idw_exit, wid_exit);
      m_dwput(idw_pktcnt, pktctr);
    
      // Check for new packets
      status = recvpacket(socket,(char *)pack_buf,sw_tl, sw_debug);

      // Output received header/data
      if (status == sw_tl) {
        pktctr++;
        m_filad(HOUT,mode_buf,mode_elem);
      }
      else if (status > 0)
        printf("Error: received < full packet -> size = %d \n",status);
      
      // Check for maximum number of output packets
      if ((pktctr_max > 0)&&(pktctr >= pktctr_max)) wid_exit = 1;
    }
    
    // Free buffers and resources
    if (pack_buf != NULL) free(pack_buf); 
    if (HOUT.open) m_close(HOUT);
    closesocket(socket,(char *)multicast_address_label.c_str(),
                       (char *)interface_address_label.c_str());
} // End: Main Routine
